-- |
-- Module    : Statistics.Constants
-- Copyright : (c) 2009, 2011 Bryan O'Sullivan
-- License   : BSD3
--
-- Maintainer  : bos@serpentine.com
-- Stability   : experimental
-- Portability : portable
--
-- Constant values common to much statistics code.
--
-- DEPRECATED: use module 'Numeric.MathFunctions.Constants' from
-- math-functions.

module Statistics.Constants
{-# DEPRECATED "use module Numeric.MathFunctions.Constants from math-functions" #-}
    ( module Numeric.MathFunctions.Constants
    ) where

import Numeric.MathFunctions.Constants
